# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation JSON Resolver module."""
